@@ -1,4 +1,4 @@
-// $Header: /cvsroot/libcwd/libcwd/include/cwd_bfd.h,v 1.18 2004/09/24 10:30:29 libcw Exp $
+// $Header: /cvsroot/libcwd/libcwd/include/cwd_bfd.h,v 1.21 2005/06/25 02:36:44 libcw Exp $
 //
 // Copyright (C) 2002 - 2004, by
 //
@@ -81,7 +81,7 @@ namespace libcwd {
 #if !CWDEBUG_LIBBFD
 typedef char* PTR;
 typedef elf32::bfd_st bfd;
-typedef elf32::Elf32_Addr bfd_vma;
+typedef char* bfd_vma;
 typedef elf32::asection_st asection;
 typedef elf32::asymbol_st asymbol;
 #endif
@@ -121,7 +121,8 @@ typedef std::list<bfile_ct*> object_files_ct;
 class bfile_ct {                                  // All allocations related to bfile_ct must be `internal'.
 private:
   bfd* M_abfd;
-  void* M_lbase;
+  void* M_lbase;	// The 'load address', or 0 for the executable.
+  void const* M_start;	// For shared libraries, this is equal to M_lbase.
   size_t M_size;
   asymbol** M_symbol_table;
   long M_number_of_symbols;
@@ -138,6 +139,7 @@ public:
 
   bfd* get_bfd(void) const { return M_abfd; }
   void* const get_lbase(void) const { return M_lbase; }
+  void const* const get_start(void) const { return M_start; }
   size_t size(void) const { return M_size; }
   asymbol** get_symbol_table(void) const { return M_symbol_table; }
   long get_number_of_symbols(void) const { return M_number_of_symbols; }
@@ -194,12 +196,13 @@ symbol_start_addr(asymbol const* s)
 // cwbfd::
 inline size_t symbol_size(asymbol const* s)
 {
-  return reinterpret_cast<size_t>(s->udata.p);
+  return static_cast<size_t>(s->udata.p);
 }
 
 // cwbfd::
 inline size_t& symbol_size(asymbol* s)
 {
+  // This assumes that sizeof(size_t) == sizeof(unsigned int).
   return *reinterpret_cast<size_t*>(&s->udata.p);
 }
 #endif
