@@ -1,4 +1,4 @@
-// $Header: /cvsroot/libcwd/libcwd/include/libcwd/bfd.h,v 1.5 2004/07/15 10:14:34 libcw Exp $
+// $Header: /cvsroot/libcwd/libcwd/include/libcwd/bfd.h,v 1.6 2005/06/29 04:10:25 libcw Exp $
 //
 // Copyright (C) 2000 - 2004, by
 // 
@@ -28,14 +28,6 @@
 #include <libcwd/pc_mangled_function_name.h>
 #endif
 
-#ifdef LIBCWD_DLOPEN_DEFINED
-#include <dlfcn.h>
-#define dlopen __libcwd_dlopen
-#define dlclose __libcwd_dlclose
-extern "C" void* dlopen(char const*, int);
-extern "C" int dlclose(void*);
-#endif // LIBCWD_DLOPEN_DEFINED
-
 #if CWDEBUG_ALLOC
 namespace libcwd {
   namespace _private_ {
