@@ -1,4 +1,4 @@
-// $Header: /cvsroot/libcwd/libcwd/include/libcwd/private_threading.h,v 1.15 2004/07/14 00:29:35 libcw Exp $
+// $Header: /cvsroot/libcwd/libcwd/include/libcwd/private_threading.h,v 1.18 2005/06/22 02:45:09 libcw Exp $
 //
 // Copyright (C) 2001 - 2004, by
 //
@@ -81,7 +81,10 @@ extern unsigned int LIBCWD_DEBUGDEBUGLOCK_CERR_count;
 #endif
 #else
 #if LIBCWD_THREAD_SAFE
-#error Fatal error: thread support was not detected during configuration of libcwd (did you use --disable-threading?)! How come you are trying to compile a threaded program now?
+#error Fatal error: thread support was not detected during configuration of libcwd (did you use --disable-threading?)! \
+       How come you are trying to compile a threaded program now? \
+       To fix this problem, either link with libcwd_r (install it), or when you are indeed compiling a \
+       single threaded application, then get rid of the -D_REENTRANT and/or -D_THREAD_SAFE in your compile flags.
 #endif
 #endif // LIBCWD_HAVE_PTHREAD
 
@@ -459,12 +462,6 @@ template <int instance>
 #else
       	PTHREAD_ADAPTIVE_MUTEX_INITIALIZER_NP;
 #endif
-
-// Specialization.
-template <>
-  extern pthread_mutex_t mutex_tct<static_tsd_instance>::S_mutex;
-      // = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
-
 #else // !LIBCWD_USE_LINUXTHREADS
       ;
 #endif // !LIBCWD_USE_LINUXTHREADS
